// RadialBasisFunction.cpp: RadialBasisFunction NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PolyTest.h"
#include "RadialBasisFunction.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

RadialBasisFunction::RadialBasisFunction()
{
	center_N = 0;
	center = NULL;
	coefficient = NULL;
	low_poly = NULL;
	isDirect = true;
}

RadialBasisFunction::~RadialBasisFunction()
{
	if(center_N != 0){
		delete[] center;
		delete[] coefficient;
		delete[] low_poly;
	}
}

float RadialBasisFunction::EvaluateFunction(float x[])
{
	if(!isDirect){
		traceTree(x, root);
		float f = evaluateRBF(x, root);
		return -(f + low_poly[0]*x[0] + low_poly[1]*x[1] + low_poly[2]*x[2] + low_poly[3]);
	}
	else{
		float x1 = x[0];
		float y1 = x[1];
		float z1 = x[2];

		float f = low_poly[0]*x1 + low_poly[1]*y1 + low_poly[2]*z1 + low_poly[3];

		for(int i=0; i<center_N; i++)
			f += coefficient[i]
				*sqrt((x1-center[i][0])*(x1-center[i][0]) 
					+ (y1-center[i][1])*(y1-center[i][1])
					+ (z1-center[i][2])*(z1-center[i][2]));

		return -f;
	}
}

void RadialBasisFunction::EvaluateGradient(float x[], float g[])
{
	g[0] = -low_poly[0];
	g[1] = -low_poly[1];
	g[2] = -low_poly[2];

	if(!isDirect){
		evaluateRBFD(x, g, root);
		g[0] *= -1.0;
		g[1] *= -1.0;
		g[2] *= -1.0;
	}
	else{
		float x1 = x[0];
		float y1 = x[1];
		float z1 = x[2];

		for(int i=0; i<center_N; i++){
			float vx = x1 - center[i][0];
			float vy = y1 - center[i][1];
			float vz = z1 - center[i][2];
			double r = sqrt(vx*vx + vy*vy + vz*vz);

			g[0] -= coefficient[i]*vx/r;
			g[1] -= coefficient[i]*vy/r;
			g[2] -= coefficient[i]*vz/r;
		}
	}
}

void RadialBasisFunction::setCenterN(int center_N)
{
	this->center_N = center_N;
	center = new double[center_N][3];
	coefficient = new double[center_N];
	low_poly = new double[4];
}

void RadialBasisFunction::setOctTree(int level)
{
	initCell(root, 0, level, origin);

	for(int i=0; i<center_N; i++)
		countCenterIndex(i, root);
	initCenterIndex(root);

	for(i=0; i<center_N; i++)
		insertCenter(i, root);

	int N = computeCellRBF(root);

	o_his = new float[level+1][3];
}

void RadialBasisFunction::initCell(RadialBasisFunction::Cell *&c, int level, int max_level, float o[3])
{
	c = new Cell;
	c->c_N = 0;
	c->level = level;
	c->size = bound_size*pow(2, -level);
	c->o[0] = o[0];
	c->o[1] = o[1];
	c->o[2] = o[2];

	if(level == max_level){
		c->isChild = true;	
		return;
	}

	c->isChild = false;

	for(int i=0; i<8; i++){
		float child_o[3];
		if(i%2 == 0)
			child_o[0] = o[0] - 0.5*c->size;
		else
			child_o[0] = o[0] + 0.5*c->size;

		if(i%4 < 2)
			child_o[1] = o[1] - 0.5*c->size;
		else
			child_o[1] = o[1] + 0.5*c->size;

		if(i < 4)
			child_o[2] = o[2] - 0.5*c->size;
		else
			child_o[2] = o[2] + 0.5*c->size;

		initCell(c->child[i], level+1, max_level, child_o);
		c->child[i]->parent = c;
	}
}

void RadialBasisFunction::setBound(float o[], float size)
{
	bound_size = size;
	origin[0] = o[0];
	origin[1] = o[1];
	origin[2] = o[2];
}

void RadialBasisFunction::countCenterIndex(int index, RadialBasisFunction::Cell *current)
{
	current->c_N++;

	if(current->isChild){
		return;
	}

	int child = 0;
	if(center[index][2] > current->o[2])
		child += 4;
	if(center[index][1] > current->o[1])
		child += 2;
	if(center[index][0] > current->o[0])
		child += 1;

	countCenterIndex(index, current->child[child]);
}

void RadialBasisFunction::initCenterIndex(RadialBasisFunction::Cell *current)
{
	current->c_index = new int[current->c_N];
	current->c_N = 0;

	if(!current->isChild)
		for(int i=0; i<8; i++)
			initCenterIndex(current->child[i]);
}

void RadialBasisFunction::insertCenter(int index, RadialBasisFunction::Cell *current)
{
	current->c_index[current->c_N++] = index;

	if(current->isChild){
		return;
	}

	int child = 0;
	if(center[index][2] > current->o[2])
		child += 4;
	if(center[index][1] > current->o[1])
		child += 2;
	if(center[index][0] > current->o[0])
		child += 1;

	insertCenter(index, current->child[child]);
}

int RadialBasisFunction::computeCellRBF(RadialBasisFunction::Cell *current)
{
	double* total = current->coeff;
	int c_N = current->c_N;
	float* o = current->o;
	
	for(int i=0; i<20; i++)
		total[i] = 0;

	for(i=0; i<c_N; i++){
		int j = current->c_index[i];
		double d = coefficient[j];
		float c[3];
		c[0] = center[j][0] - o[0];
		c[1] = center[j][1] - o[1];
		c[2] = center[j][2] - o[2];

		total[0] += d;

		total[1] += -d*c[0];
		total[2] += -d*c[1];
		total[3] += -d*c[2];

		total[4] += 0.5*d*(c[1]*c[1] + c[2]*c[2]);
		total[5] += 0.5*d*(c[2]*c[2] + c[0]*c[0]);
		total[6] += 0.5*d*(c[0]*c[0] + c[1]*c[1]);
		total[7] += -d*c[0]*c[1];
		total[8] += -d*c[1]*c[2];
		total[9] += -d*c[2]*c[0];

		total[10] += 0.5*d*c[0]*(c[1]*c[1] + c[2]*c[2]);
		total[11] += 0.5*d*c[1]*(c[2]*c[2] + c[0]*c[0]);
		total[12] += 0.5*d*c[2]*(c[0]*c[0] + c[1]*c[1]);
		total[13] += 0.5*d*c[1]*(c[1] + c[2]*c[2] - 2.0*c[0]*c[0]);
		total[14] += 0.5*d*c[2]*(c[2] + c[1]*c[1] - 2.0*c[0]*c[0]);
		total[15] += 0.5*d*c[0]*(c[0] + c[2]*c[2] - 2.0*c[1]*c[1]);
		total[16] += 0.5*d*c[2]*(c[2] + c[0]*c[0] - 2.0*c[1]*c[1]);
		total[17] += 0.5*d*c[0]*(c[0] + c[1]*c[1] - 2.0*c[2]*c[2]);
		total[18] += 0.5*d*c[1]*(c[1] + c[0]*c[0] - 2.0*c[2]*c[2]);
		total[19] += -3.0*d*c[0]*c[1]*c[2];
	}
	if(!current->isChild)
		for(i=0; i<8; i++)
			computeCellRBF(current->child[i]);
	return 0;
}

float RadialBasisFunction::evaluateRBF(float x[], RadialBasisFunction::Cell *current)
{
	float f = 0;
	if(current->isChild){
		for(int i=0; i<current->c_N; i++){
			int j = current->c_index[i];
			f += coefficient[j]
			*sqrt((x[0]-center[j][0])*(x[0]-center[j][0]) 
				+ (x[1]-center[j][1])*(x[1]-center[j][1])
				+ (x[2]-center[j][2])*(x[2]-center[j][2]));
		}
	}
	else{
		int l = current->level+1;
		float* o = o_his[l];
		float T = current->size*1.5;//15*current->size*current->size*0.25;

		for(int i=0; i<8; i++){
			Cell* c = current->child[i];
			
			if(c->c_N == 0)
				continue;

			if(T < max(fabs(c->o[0] - o[0]), max(fabs(c->o[1] - o[1]), fabs(c->o[2] - o[2])))){
				double* d = c->coeff;
				float x1 = x[0] - c->o[0];
				float y1 = x[1] - c->o[1];
				float z1 = x[2] - c->o[2];
				double n = sqrt(x1*x1 + y1*y1 + z1*z1); 
				f += d[0]*n + (d[1]*x1 + d[2]*y1 + d[3]*z1)/n + 
					(d[4]*x1*x1 + d[5]*y1*y1 + d[6]*z1*z1 + 
					d[7]*x1*y1 + d[8]*y1*z1 + d[9]*z1*x1)/pow(n, 3.0); /* +
					(d[10]*x1*x1*x1 + d[11]*y1*y1*y1 + d[12]*z1*z1*z1
					+ d[13]*x1*x1*y1 + d[14]*x1*x1*z1 
					+ d[15]*y1*y1*x1 + d[16]*y1*y1*z1
					+ d[17]*z1*z1*x1 + d[18]*z1*z1*y1 + d[19]*x1*y1*z1)/pow(n, 5.0);*/
			}
			else{
				f += evaluateRBF(x, c);
			}
		}
	}
	return f;
}

void RadialBasisFunction::traceTree(float x[], RadialBasisFunction::Cell *current)
{
	float *o = current->o;
	int l = current->level;
	o_his[l][0] = o[0];
	o_his[l][1] = o[1];
	o_his[l][2] = o[2];

	if(current->isChild)
		return;

	int child = 0;
	if(x[2] > o[2])
		child += 4;
	if(x[1] > o[1])
		child += 2;
	if(x[0] > o[0])
		child += 1;

	traceTree(x, current->child[child]);
}

void RadialBasisFunction::evaluateRBFD(float x[], float* g, RadialBasisFunction::Cell *current)
{
	if(current->isChild){
		for(int i=0; i<current->c_N; i++){
			int j = current->c_index[i];
			float vx = x[0] - center[j][0];
			float vy = x[1] - center[j][1];
			float vz = x[2] - center[j][2];
			double r = sqrt(vx*vx + vy*vy + vz*vz);
			g[0] += coefficient[j]*vx/r;
			g[1] += coefficient[j]*vy/r;
			g[2] += coefficient[j]*vz/r;
		}
	}
	else{
		int l = current->level+1;
		float* o = o_his[l];
		float T = current->size*1.5;

		for(int i=0; i<8; i++){
			Cell* c = current->child[i];
			
			if(c->c_N == 0)
				continue;

			if(T < max(fabs(c->o[0] - o[0]), max(fabs(c->o[1] - o[1]), fabs(c->o[2] - o[2])))){
				double* d = c->coeff;
				float x1 = x[0] - c->o[0];
				float y1 = x[1] - c->o[1];
				float z1 = x[2] - c->o[2];
				double n = sqrt(x1*x1 + y1*y1 + z1*z1); 

				g[0] += d[0]*x1/n 
					+ (d[1]*(y1*y1 + z1*z1) - d[2]*x1*y1 - d[3]*x1*z1)/pow(n, 3.0);
				/*
					+ (d[4]*x1*(2.0*(y1*y1 + z1*z1) - x1*x1) 
					 - d[5]*3.0*x1*y1*y1 - d[6]*3.0*x1*z1*z1 
					 + d[7]*y1*(-2.0*x1*x1 + y1*y1 + z1*z1)  
					 - d[8]*3.0*x1*y1*z1 
					 + d[9]*z1*(-2.0*x1*x1 + y1*y1 + z1*z1))/pow(n, 5.0);*/

				g[1] += d[0]*y1/n
					+ (-d[1]*x1*y1 + d[2]*(x1*x1 + z1*z1) - d[3]*y1*z1)/pow(n, 3.0);
				/*
					+ (d[4]*(-3.0)*x1*x1*y1 
					 + d[5]*y1*(2.0*(x1*x1 + z1*z1) - y1*y1)
					 + d[6]*(-3.0)*y1*z1*z1
					 + d[7]*x1*(x1*x1 - 2.0*y1*y1 + z1*z1)
					 + d[8]*z1*(x1*x1 - 2.0*y1*y1 + z1*z1)
					 - d[9]*3.0*x1*y1*z1)/pow(n, 5.0);*/

				g[2] +=  d[0]*z1/n
					+ (-d[1]*y1*z1 - d[2]*y1*z1 + d[3]*(x1*x1 + y1*y1))/pow(n, 3.0);
				/*
					+ (d[4]*(-3.0)*x1*x1*z1
					 + d[5]*(-3.0)*y1*y1*z1
					 + d[6]*z1*(2.0*(x1*x1 + y1*y1) - z1*z1)
					 - d[7]*3.0*x1*y1*z1
					 + d[8]*y1*(x1*x1 + y1*y1 - 2.0*z1*z1)
					 + d[9]*x1*(x1*x1 + y1*y1 - 2.0*z1*z1))/pow(n, 5.0);*/
			}
			else{
				evaluateRBFD(x, g, c);
			}
		}
	}
}
