// Smoother.h: Smoother NX̃C^[tFCX
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SMOOTHER_H__E5F4FBA7_6B18_4975_A8CF_6D3D89B26D41__INCLUDED_)
#define AFX_SMOOTHER_H__E5F4FBA7_6B18_4975_A8CF_6D3D89B26D41__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "MeshData.h"
#include "vtkImplicitFunction.h"
#include "OurImplicitFunction.h"
#include "Node.h"

#define EPSILON 0.0000000001

class Smoother  
{
public:
	MeshData* mesh;
	float (*T)[3];
	BOOL isEngrave;
	float (*center)[3];
	float (*m)[3];
	BOOL *valid;

protected:
	void detectSolidBound();
	BOOL *isSolidBound;
	int *v_table;
	float (*originalNormal)[3];
	int table_N;

public:
	void setNormalAsGradWithoutNormalize(vtkImplicitFunction *func);
	void correctImplicitSurfaceNew(vtkImplicitFunction *func, int iter, int off_iter, BOOL error);
	float computeMaxFaceNormalErr(vtkImplicitFunction *func);
	float computeMaxVertexErr(vtkImplicitFunction *func);
	void moveNormalD2(int iter, float dt);
	void moveVNormal(float dt);
	void moveNormalWithF(float C, vtkImplicitFunction* f);
	void moveNormalD(int iter, float dt);
	BOOL searchNearPoint2dire(vtkImplicitFunction* func, float start[], float p[], float interval, int MAX);
	BOOL searchNearPoint1D(vtkImplicitFunction* func, float start[], float p[], float *d, float interval, int MAX);
	void addVolumeW2(double (*Q)[10], float C);
	void addVolumeW(double (*Q)[10], float C);
	void minimizeH1(int iter, double (*Q)[10], float C);
	void addDerivativeError(double (*Q)[10],  float C);
	void addUniformW2(double (*Q)[10], double (*dir)[3], float C);
	void minimizeQPowell(double (*Q)[10], float T);
	void addUniformW(double (*Q)[10], double (*dir)[3], float C);
	void computeDir(double (*dir)[3]);
	void minimizeQ1D(double (*Q)[10], double (*d)[3]);
	void replaceOptimalPosition(int MAX, float uni_w, int ave);
	void generateUniformCenter(vtkImplicitFunction* func, int iter, float W);
	void checkFlip(vtkImplicitFunction* func);
	void averagingQ(int iter, double (*Q)[10]);
	void minimizeQ(int iter, double (*Q)[10]);
	void computeQuadric(double (*Q)[10]);
	void resolveFlip();
	void setVertexFromCenterW(float T);
	void fillValid();
	void uniformizeCenter(float C);
	void setCenterOnSurface(vtkImplicitFunction* func);
	BOOL checkZone(int i, float opt[3]);
	BOOL computeOptP(double Q[10], float p[3], float opt[3], float C);
	BOOL computeFaceQ(int i, vtkImplicitFunction* func, double Q[10]);
	BOOL searchNearPoint(vtkImplicitFunction* func, float start[3], float p[3], float interval, int MAX);
	void optimizeSquentily(vtkImplicitFunction *func, float T);
	void computeVertexBisection(float interval, vtkImplicitFunction* func, int MAX);
	void setOpt();
	void moveNormalCenter2();
	void setCenterNormal(vtkImplicitFunction* func);
	void setVertexFromCenter();
	void computeCenterBisection(float interval, vtkImplicitFunction* func, int MAX);
	void correctImplicitSurfacePowell(OurImplicitFunction *func);
	void keepOnSurfaceUTN(vtkImplicitFunction *func, float c1, float c2, float c3, float T);
	void keepOnSurfaceUT(vtkImplicitFunction *func, float c1, float c2, float T);
	void keepOnSurfaceT(vtkImplicitFunction *func, float c);
	void offset(vtkImplicitFunction* func);
	void LapMesh(float dt);
	void moveImplicit(vtkImplicitFunction *func, int iter, BOOL isEngrave);
	void tangentLapMesh(float dt);
	void keepOnSurfaceMesh(vtkImplicitFunction *func);
	void moveNormalMesh(vtkImplicitFunction* func);
	void setNormalAsGraCeneter(vtkImplicitFunction* func);
	void moveNormalCenter();
	void correntImplicitCenter(vtkImplicitFunction* func, int iter);
	void tangentLapCenter(float dt);
	void keepOnSurfaceCenter(vtkImplicitFunction* func);
	void setFixedVertex(float T, vtkImplicitFunction* func);
	void moveNormal2(float (*F)[3]);
	void keepOnSurface2(vtkImplicitFunction *func, float (*F)[3]);
	void tangentLap2(float dt, float (*F)[3]);
	void correctImplicitSurface2(vtkImplicitFunction *func, int iter, float T, int projection, int uniform, int keep, BOOL err);
	void setVertexNormalAsGrad(vtkImplicitFunction* func);
	float computeFaceNormalErr(vtkImplicitFunction *func);
	float computeVertexErr(vtkImplicitFunction* func);
	void tangentLapW(float dt, float T);
	void setNormalAsGrad(vtkImplicitFunction* func);
	void principalFlow(float dt, float T);
	void tangentLap(float dt, int from);
	void keepOnSurface(vtkImplicitFunction* func, int from);
	void moveNormal(int from);
	void smoothNormal(float T);
	void correctImplicitSurface(vtkImplicitFunction* func, int iter, int off_iter, float T, int projection, int uniform, int keep, BOOL err, int from);
	void meanCurvatureFlow12Ring(int iter, float dt, float e);
	void computeNRingMeanCurvature(int n);
	void GaussianNoise(float size);
	void setMeshData(MeshData* mesh);
	void meanCurvatureFlow(int iter, float dt);
	void LaplacianFlow(int iter, float dt);
	Smoother();
	virtual ~Smoother();

	static inline BOOL INVERSE(double B[10], double A[10]){
		double d = DET(A);
		if(fabs(d) < EPSILON)
			return false;
		B[0] = (A[3]*A[5] - A[4]*A[4])/d;
		B[1] = (A[2]*A[4] - A[1]*A[5])/d;
		B[2] = (A[1]*A[4] - A[2]*A[3])/d;
		B[3] = (A[0]*A[5] - A[2]*A[2])/d;
		B[4] = (A[1]*A[2] - A[0]*A[4])/d;
		B[5] = (A[0]*A[3] - A[1]*A[1])/d;
		return true;
	}

	static inline double DET(double A[10]){
		return A[0]*A[3]*A[5] + 2.0*A[1]*A[4]*A[2] 
			-A[2]*A[2]*A[3] - A[1]*A[1]*A[5] - A[4]*A[4]*A[0];
	}

	static inline void MATRIX(double A[10], double n[3], double d){
		A[0] = n[0]*n[0];
		A[1] = n[0]*n[1];
		A[2] = n[0]*n[2];
		A[3] = n[1]*n[1];
		A[4] = n[1]*n[2];
		A[5] = n[2]*n[2];
		A[6] = d*n[0];
		A[7] = d*n[1];
		A[8] = d*n[2];
		A[9] = d*d;
	}

	static inline void MAT_TIMES(double A[10], double k){
		A[0] *= k;
		A[1] *= k;
		A[2] *= k;
		A[3] *= k;
		A[4] *= k;
		A[5] *= k;
		A[6] *= k;
		A[7] *= k;
		A[8] *= k;
		A[9] *= k;
	}

	static inline void MAT_SUM(double B[6], double A[6]){
		B[0] += A[0];
		B[1] += A[1];
		B[2] += A[2];
		B[3] += A[3];
		B[4] += A[4];
		B[5] += A[5];
		B[6] += A[6];
		B[7] += A[7];
		B[8] += A[8];
		B[9] += A[9];
	}

	static inline void MAT_BY_VEC(double v[3], double A[10], double b[3]){
		v[0] = A[0]*b[0] + A[1]*b[1] + A[2]*b[2];
		v[1] = A[1]*b[0] + A[3]*b[1] + A[4]*b[2];
		v[2] = A[2]*b[0] + A[4]*b[1] + A[5]*b[2];
	}

	static inline void MAT_INIT(double A[10]){
		A[0] = A[1] = A[2] = A[3] = A[4] = A[5] = A[6] = A[7] = A[8] = A[9] = 0;
	}

	static inline void MAT_PLUS(double C[10], double A[10], double B[10]){
		C[0] = A[0] + B[0];
		C[1] = A[1] + B[1];
		C[2] = A[2] + B[2];
		C[3] = A[3] + B[3];
		C[4] = A[4] + B[4];
		C[5] = A[5] + B[5];
		C[6] = A[6] + B[6];
		C[7] = A[7] + B[7];
		C[8] = A[8] + B[8];
		C[9] = A[9] + B[9];
	}

	static inline void MAT_COPY(double A[10], double B[10]){
		A[0] = B[0];
		A[1] = B[1];
		A[2] = B[2];
		A[3] = B[3];
		A[4] = B[4];
		A[5] = B[5];
		A[6] = B[6];
		A[7] = B[7];
		A[8] = B[8];
		A[9] = B[9];
	}

	static inline double Q_ERR(double A[10], float v[3]){
		return v[0]*(A[0]*v[0] + A[1]*v[1] + A[2]*v[2]) +
			   v[1]*(A[1]*v[0] + A[3]*v[1] + A[4]*v[2]) +
			   v[2]*(A[2]*v[0] + A[4]*v[1] + A[5]*v[2]) +
			   2.0*(A[6]*v[0] + A[7]*v[1] + A[8]*v[2]) + 
			   A[9];
	}

	static inline double Q_ERR(double A[10], double v[3]){
		return v[0]*(A[0]*v[0] + A[1]*v[1] + A[2]*v[2]) +
			   v[1]*(A[1]*v[0] + A[3]*v[1] + A[4]*v[2]) +
			   v[2]*(A[2]*v[0] + A[4]*v[1] + A[5]*v[2]) +
			   2.0*(A[6]*v[0] + A[7]*v[1] + A[8]*v[2]) + 
			   A[9];
	}

	static inline void upheap(float *a, int N, int k, int *p, int *q){
		int v;
		v = p[k];
		while(k > 1 && a[p[k/2]] >= a[v]){
			p[k] = p[k/2]; q[p[k/2]] = k; k = k/2;
		}
		p[k] = v; q[v] = k;
	}

	inline void downheap(float *a, int N, int k, int *p, int *q){
		int j, v;
		v = p[k];
		while(k <= N/2){
			j = k+k;
			if(j < N && a[p[j]] > a[p[j+1]]) j++;
			if(a[v] <= a[p[j]]) break;
			p[k] = p[j]; q[p[j]] = k; k = j;
		}
		p[k] = v; q[v] = k;
	}

	static inline float SHAPE(float p1[3], float p2[3], float p3[3]){
		double l1 = (p1[0]-p2[0])*(p1[0]-p2[0])+(p1[1]-p2[1])*(p1[1]-p2[1])+(p1[2]-p2[2])*(p1[2]-p2[2]);
		double l2 = (p2[0]-p3[0])*(p2[0]-p3[0])+(p2[1]-p3[1])*(p2[1]-p3[1])+(p2[2]-p3[2])*(p2[2]-p3[2]);
		double l3 = (p3[0]-p1[0])*(p3[0]-p1[0])+(p3[1]-p1[1])*(p3[1]-p1[1])+(p3[2]-p1[2])*(p3[2]-p1[2]);
		double area = MeshData::AREA(p1, p2, p3);
		return (float)(4.0*sqrt(3.0)*area/(l1 + l2 + l3));
	}
};

#endif // !defined(AFX_SMOOTHER_H__E5F4FBA7_6B18_4975_A8CF_6D3D89B26D41__INCLUDED_)
