// VolumeFunction.cpp: VolumeFunction NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PolyTest.h"
#include "VolumeFunction.h"
#include "vtkStructuredPoints.h"
#include "MeshData.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define E 2.7182818284590

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

VolumeFunction::VolumeFunction()
{
	value = NULL;
}

VolumeFunction::~VolumeFunction()
{
	//if(value != NULL)
		//delete value;
}

void VolumeFunction::EvaluateGradient(float x[], float g[])
{
	float X = x[0] - originX;
	float Y = x[1] - originY;
	float Z = x[2] - originZ;
	
	if(X < 0)
		X = 0;
	else if(X > spaceX*(dimX-2))
		X = spaceX*(dimX-2);

	if(Y < 0)
		Y = 0;
	else if(Y > spaceY*(dimY-2))
		Y = spaceY*(dimY-2);

	if(Z < 0)
		Z = 0;
	else if(Z > spaceZ*(dimZ-2))
		Z = spaceZ*(dimZ-2);

	/*
	if(X < 1)
		X = 1;
	else if(X > spaceX*(dimX-3))
		X = spaceX*(dimX-3);

	if(Y < 1)
		Y = 1;
	else if(Y > spaceY*(dimY-3))
		Y = spaceY*(dimY-3);

	if(Z < 1)
		Z = 1;
	else if(Z > spaceZ*(dimZ-3))
		Z = spaceZ*(dimZ-3);
		*/

	int xi = (int)(X/spaceX);
	int yi = (int)(Y/spaceY);
	int zi = (int)(Z/spaceZ);

	float fx = (X - (float)xi*spaceX)/spaceX;
	float fy = (Y - (float)yi*spaceY)/spaceY;
	float fz = (Z - (float)zi*spaceZ)/spaceZ;

	float* v000 = nabla[zi][yi][xi];
	//this->computGradientAtP(xi, yi, zi, v000);
	float* v001 = nabla[zi+1][yi][xi];
	//this->computGradientAtP(xi+1, yi, zi, v001);
	float* v010 = nabla[zi][yi+1][xi];
	//this->computGradientAtP(xi, yi+1, zi, v010);
	float* v011 = nabla[zi+1][yi+1][xi];
	//this->computGradientAtP(xi+1, yi+1, zi, v011);
	float* v100 = nabla[zi][yi][xi+1];
	//this->computGradientAtP(xi, yi, zi+1, v100);
	float* v101 = nabla[zi+1][yi][xi+1];
	//this->computGradientAtP(xi+1, yi, zi+1, v101);
	float* v110 = nabla[zi][yi+1][xi+1];
	//this->computGradientAtP(xi, yi+1, zi+1, v110);
	float* v111 = nabla[zi+1][yi+1][xi+1];
	//this->computGradientAtP(xi+1, yi+1, zi+1, v111);
	
	float vx00[3];
	INTER(vx00, v000, v100, fx);
	float vx01[3];
	INTER(vx01, v001, v101, fx);
	float vx10[3];
	INTER(vx10, v010, v110, fx);
	float vx11[3];
	INTER(vx11, v011, v111, fx);

	float vxy0[3];
	INTER(vxy0, vx00, vx10, fy);
	float vxy1[3];
	INTER(vxy1, vx01, vx11, fy);

	INTER(g, vxy0, vxy1, fz);
}

float VolumeFunction::EvaluateFunction(float x[])
{
	float X = x[0] - originX;
	float Y = x[1] - originY;
	float Z = x[2] - originZ;
	
	if(X < 0)
		X = 0;
	else if(X > spaceX*(dimX-2))
		X = spaceX*(dimX-2);

	if(Y < 0)
		Y = 0;
	else if(Y > spaceY*(dimY-2))
		Y = spaceY*(dimY-2);

	if(Z < 0)
		Z = 0;
	else if(Z > spaceZ*(dimZ-2))
		Z = spaceZ*(dimZ-2);

	int xi = (int)(X/spaceX);
	int yi = (int)(Y/spaceY);
	int zi = (int)(Z/spaceZ);

	float fx = (X - (float)xi*spaceX)/spaceX;
	float fy = (Y - (float)yi*spaceY)/spaceY;
	float fz = (Z - (float)zi*spaceZ)/spaceZ;

	/*
	float v000 = value[zi][yi][xi];
	float v001 = value[zi][yi][xi+1];
	float v010 = value[zi][yi+1][xi];
	float v011 = value[zi][yi+1][xi+1];
	float v100 = value[zi+1][yi][xi];
	float v101 = value[zi+1][yi][xi+1];
	float v110 = value[zi+1][yi+1][xi];
	float v111 = value[zi+1][yi+1][xi+1];
	*/
	float v000 = value[zi][yi][xi];
	float v001 = value[zi+1][yi][xi];
	float v010 = value[zi][yi+1][xi];
	float v011 = value[zi+1][yi+1][xi];
	float v100 = value[zi][yi][xi+1];
	float v101 = value[zi+1][yi][xi+1];
	float v110 = value[zi][yi+1][xi+1];
	float v111 = value[zi+1][yi+1][xi+1];

	float vx00 = v000 + fx*(v100 - v000);
	float vx01 = v001 + fx*(v101 - v001);
	float vx10 = v010 + fx*(v110 - v010);
	float vx11 = v011 + fx*(v111 - v011);

	float vxy0 = vx00 + fy*(vx10 - vx00);
	float vxy1 = vx01 + fy*(vx11 - vx01);

	return vxy0 + fz*(vxy1 - vxy0);
}

void VolumeFunction::setDim(int dim[])
{
	dimX = dim[0];
	dimY = dim[1];
	dimZ = dim[2];
	value = new float**[dimZ+2*mask_size+2];
	for(int i=0; i<dimZ+2*mask_size+2; i++){
		value[i] = new float*[dimY+2*mask_size+2];
		for(int j=0; j<dimY+2*mask_size+2; j++){
			value[i][j] = new float[dimX+2*mask_size+2];
			value[i][j] += mask_size+1;
		}
		value[i] += mask_size+1;
	}
	value += mask_size+1;
}

void VolumeFunction::setOriAndSpace(float origin[], float space[])
{
	spaceX = space[0];
	spaceY = space[1];
	spaceZ = space[2];

	originX = origin[0];
	originY = origin[1];
	originZ = origin[2];
}



void VolumeFunction::computGradientAtP(int ix, int iy, int iz, float g[])
{
	g[0] = g[1] = g[2] = 0;
	double total = 0;
	for(int i=-mask_size; i<=mask_size; i++)
		for(int j=-mask_size; j<=mask_size; j++)
			for(int k=-mask_size; k<=mask_size; k++){
				float m = mask[i]*mask[k]*mask[j]*value[iz+i][iy+j][ix+k];
				g[0] += (float)k*m;
				g[1] += (float)j*m;
				g[2] += (float)i*m;
			}
	g[0] /= -spaceX;
	g[1] /= -spaceY;
	g[2] /= -spaceZ;

	/*
	g[0] = (8*(value[iz][iy][ix+1] - value[iz][iy][ix-1]) 
			+ 2*(value[iz][iy+1][ix+1] + value[iz][iy-1][ix+1] 
				- value[iz][iy+1][ix-1] - value[iz][iy-1][ix-1]
				+ value[iz+1][iy][ix+1] + value[iz-1][iy][ix+1] 
				- value[iz+1][iy][ix-1] - value[iz-1][iy][ix-1])
			+ (value[iz+1][iy+1][ix+1] + value[iz+1][iy-1][ix+1] 
				- value[iz+1][iy+1][ix-1] - value[iz+1][iy-1][ix-1]
				+ value[iz-1][iy+1][ix+1] + value[iz-1][iy-1][ix+1] 
				- value[iz-1][iy+1][ix-1] - value[iz-1][iy-1][ix-1]))/(spaceX); 
	
	
	g[1] = (8*(value[iz][iy+1][ix] - value[iz][iy-1][ix])
			+ 2*(value[iz][iy+1][ix+1] + value[iz][iy+1][ix-1] 
				- value[iz][iy-1][ix+1] - value[iz][iy-1][ix-1]
				+ value[iz+1][iy+1][ix] + value[iz-1][iy+1][ix] 
				- value[iz+1][iy-1][ix] - value[iz-1][iy-1][ix])
			+ (value[iz+1][iy+1][ix+1] + value[iz+1][iy+1][ix-1] 
				- value[iz+1][iy-1][ix+1] - value[iz+1][iy-1][ix-1]
				+ value[iz-1][iy+1][ix+1] + value[iz-1][iy+1][ix-1] 
				- value[iz-1][iy-1][ix+1] - value[iz-1][iy-1][ix-1]))/(spaceY);
				
	g[2] = (8*(value[iz+1][iy][ix] - value[iz-1][iy][ix])  
		     + 2*(value[iz+1][iy][ix+1] + value[iz+1][iy][ix-1] 
				- value[iz-1][iy][ix+1] - value[iz-1][iy][ix-1]
				+ value[iz+1][iy+1][ix] + value[iz+1][iy-1][ix] 
				- value[iz-1][iy+1][ix] - value[iz-1][iy-1][ix])
			+ (value[iz+1][iy+1][ix+1] + value[iz+1][iy+1][ix-1] 
				- value[iz-1][iy+1][ix+1] - value[iz-1][iy+1][ix-1]
				+ value[iz+1][iy-1][ix+1] + value[iz+1][iy-1][ix-1] 
				- value[iz-1][iy-1][ix+1] - value[iz-1][iy-1][ix-1]))/(spaceZ);
				*/
			
		
	//g[1] = (value[iz][iy+1][ix] - value[iz][iy-1][ix])/(2.0f*spaceY);
	//g[2] = (value[iz+1][iy][ix] - value[iz-1][iy][ix])/(2.0f*spaceZ);
}

void VolumeFunction::setMaskSize(float T)
{
	mask_size = (int)4*T + 1;
	mask = new double[mask_size*2+1];
	mask += mask_size;

	double total = 0;
	for(int i=-mask_size; i<=mask_size; i++){
		if (fabs(i) <= 2*T)
			mask[i] = exp(-(i*i)/(2*T*T));
		else if (2*T < fabs(i) && fabs(i) <= 4*T)
			mask[i] = 1/(16*E*E) * pow(4-fabs(i)/T, 4);
		else
			mask[i] = 0;
		total += mask[i];
	}
	for(i=-mask_size; i<=mask_size; i++)
		mask[i] /= total;

}

void VolumeFunction::reflectData()
{
	for(int i=-mask_size-1; i<dimZ+mask_size+1; i++)
		for(int j=-mask_size-1; j<dimY+mask_size+1; j++)
			for(int k=-mask_size-1; k<dimX+mask_size+1; k++){
				bool flag = false;
				int X = k;
				int Y = j;
				int Z = i;
				if(i < 0){
					flag = true;
					Z = -i;
				}
				else if(i >= dimZ){
					flag = true;
					Z = - i + 2*dimZ - 2;
				}
				if(j < 0){
					flag = true;
					Y = -j;
				}
				else if(j >= dimY){
					flag = true;
					Y = - j + 2*dimY - 2;
				}
				if(k < 0){
					flag = true;
					X = -k;
				}
				else if(k >= dimX){
					flag = true;
					X = - k + 2*dimX - 2;
				}
				if(flag)
					value[i][j][k] = value[Z][Y][X];
			}
}

void VolumeFunction::computeNabla()
{
	nabla = (float(***)[3])new float**[dimZ];
	for(int i=0; i<dimZ; i++){
		nabla[i] = (float(**)[3])new float*[dimY];
		for(int j=0; j<dimY; j++){
			nabla[i][j] = new float[dimX][3];
		}
	}

	float*** I = new float**[dimZ+2*mask_size+2];
	for(i=0; i<dimZ+2*mask_size+2; i++){
		I[i] = new float*[dimY+2*mask_size+2];
		for(int j=0; j<dimY+2*mask_size+2; j++){
			I[i][j] = new float[dimX+2*mask_size+2];
			I[i][j] += mask_size+1;
		}
		I[i] += mask_size+1;
	}
	I += mask_size+1;

	for(i=-mask_size; i<dimZ+mask_size; i++)
		for(int j=-mask_size; j<dimY+mask_size; j++)
			for(int k=-mask_size; k<dimX+mask_size; k++)
				I[i][j][k] = value[i][j][k];

	convoluteY(I);
	convoluteZ(I);
	computeNablaX(I);

	for(i=-mask_size; i<dimZ+mask_size; i++)
		for(int j=-mask_size; j<dimY+mask_size; j++)
			for(int k=-mask_size; k<dimX+mask_size; k++)
				I[i][j][k] = value[i][j][k];

	convoluteZ(I);
	convoluteX(I);
	computeNablaY(I);

	for(i=-mask_size; i<dimZ+mask_size; i++)
		for(int j=-mask_size; j<dimY+mask_size; j++)
			for(int k=-mask_size; k<dimX+mask_size; k++)
				I[i][j][k] = value[i][j][k];

	convoluteX(I);
	convoluteY(I);
	computeNablaZ(I);

	I -= mask_size+1;
	for(i=0; i<dimZ+2*mask_size+2; i++){
		I[i] -= mask_size+1;
		for(int j=0; j<dimY+2*mask_size+2; j++){
			I[i][j] -= mask_size+1;
			delete[] I[i][j];
		}
		delete[] I[i];
	}
	delete[] I;
}

void VolumeFunction::convoluteX(float ***I)
{
	float *tmp = new float[dimX];
	for(int i=-mask_size; i<dimZ+mask_size; i++){
		for(int j=-mask_size; j<dimY+mask_size; j++){	
			for(int k=0; k<dimX; k++){
				tmp[k] = 0;
				for(int m=-mask_size; m<=mask_size; m++){
					tmp[k] += mask[m]*I[i][j][k+m];
				}
			}
			for(k=0; k<dimX; k++)
				I[i][j][k] = tmp[k];
		}
	}
	delete[] tmp;
}

void VolumeFunction::convoluteY(float ***I)
{
	float *tmp = new float[dimY];
	for(int i=-mask_size; i<dimX+mask_size; i++){
		for(int j=-mask_size; j<dimZ+mask_size; j++){	
			for(int k=0; k<dimY; k++){
				tmp[k] = 0;
				for(int m=-mask_size; m<=mask_size; m++){
					tmp[k] += mask[m]*I[j][k+m][i];
				}
			}
			for(k=0; k<dimY; k++)
				I[j][k][i] = tmp[k];
		}
	}
	delete[] tmp;
}

void VolumeFunction::convoluteZ(float ***I)
{
	float *tmp = new float[dimZ];
	for(int i=-mask_size; i<dimY+mask_size; i++){
		for(int j=-mask_size; j<dimX+mask_size; j++){	
			for(int k=0; k<dimZ; k++){
				tmp[k] = 0;
				for(int m=-mask_size; m<=mask_size; m++){
					tmp[k] += mask[m]*I[k+m][i][j];
				}
			}
			for(k=0; k<dimZ; k++)
				I[k][i][j] = tmp[k];
		}
	}
	delete[] tmp;
}

void VolumeFunction::computeNablaX(float ***I)
{
	for(int i=0; i<dimZ; i++){
		for(int j=0; j<dimY; j++){	
			for(int k=0; k<dimX; k++){
				float tmp = 0;
				for(int m=-mask_size; m<=mask_size; m++){
					tmp += m*mask[m]*I[i][j][k+m];
				}
				nabla[i][j][k][0] = tmp/spaceX;
			}
		}
	}
}

void VolumeFunction::computeNablaY(float ***I)
{
	for(int i=0; i<dimX; i++){
		for(int j=0; j<dimZ; j++){	
			for(int k=0; k<dimY; k++){
				float tmp = 0;
				for(int m=-mask_size; m<=mask_size; m++){
					tmp += m*mask[m]*I[j][k+m][i];
				}
				nabla[j][k][i][1] = tmp/spaceY;
			}
		}
	}
}

void VolumeFunction::computeNablaZ(float ***I)
{
	for(int i=0; i<dimY; i++){
		for(int j=0; j<dimX; j++){	
			for(int k=0; k<dimZ; k++){
				float tmp = 0;
				for(int m=-mask_size; m<=mask_size; m++){
					tmp += m*mask[m]*I[k+m][i][j];
				}
				nabla[k][i][j][2] = tmp/spaceZ;
			}
		}
	}
}

void VolumeFunction::GaussSmoothing()
{
	float*** I = new float**[dimZ+2*mask_size+2];
	for(int i=0; i<dimZ+2*mask_size+2; i++){
		I[i] = new float*[dimY+2*mask_size+2];
		for(int j=0; j<dimY+2*mask_size+2; j++){
			I[i][j] = new float[dimX+2*mask_size+2];
			I[i][j] += mask_size+1;
		}
		I[i] += mask_size+1;
	}
	I += mask_size+1;

	for(i=-mask_size; i<dimZ+mask_size; i++)
		for(int j=-mask_size; j<dimY+mask_size; j++)
			for(int k=-mask_size; k<dimX+mask_size; k++)
				I[i][j][k] = value[i][j][k];

	convoluteX(I);
	convoluteY(I);
	convoluteZ(I);

	for(i=-mask_size; i<dimZ+mask_size; i++)
		for(int j=-mask_size; j<dimY+mask_size; j++)
			for(int k=-mask_size; k<dimX+mask_size; k++)
				value[i][j][k] = I[i][j][k];

	I -= mask_size+1;
	for(i=0; i<dimZ+2*mask_size+2; i++){
		I[i] -= mask_size+1;
		for(int j=0; j<dimY+2*mask_size+2; j++){
			I[i][j] -= mask_size+1;
			delete[] I[i][j];
		}
		delete[] I[i];
	}
	delete[] I;
}

void VolumeFunction::thresholding()
{
	for(int i=-mask_size-1; i<dimZ+mask_size+1; i++)
		for(int j=-mask_size-1; j<dimY+mask_size+1; j++)
			for(int k=-mask_size-1; k<dimX+mask_size+1; k++)
				if(value[i][j][k] == 0)
					value[i][j][k] = 125;
				else
					value[i][j][k] = -125;
}

void VolumeFunction::computeSimpleGradient()
{
	nabla = (float(***)[3])new float**[dimZ];
	for(int i=0; i<dimZ; i++){
		nabla[i] = (float(**)[3])new float*[dimY];
		for(int j=0; j<dimY; j++){
			nabla[i][j] = new float[dimX][3];
		}
	}

	for(i=0; i<dimZ; i++){
		for(int j=0; j<dimY; j++){
			for(int k=0; k<dimX; k++){
				nabla[i][j][k][0] = (value[i][j][k+1] - value[i][j][k-1])/(2.0f*spaceX);
				nabla[i][j][k][1] = (value[i][j+1][k] - value[i][j-1][k])/(2.0f*spaceY);
				nabla[i][j][k][2] = (value[i+1][j][k] - value[i-1][j][k])/(2.0f*spaceZ);
			}
		}
	}
}

void VolumeFunction::averageGradient(int iter, float T)
{
	float (***tmp)[3] = (float(***)[3])new float**[dimZ];
	for(int i=0; i<dimZ; i++){
		tmp[i] = (float(**)[3])new float*[dimY];
		for(int j=0; j<dimY; j++){
			tmp[i][j] = new float[dimX][3];
		}
	}

	for(int m=0; m<iter; m++){
		for(i=1; i<dimZ-1; i++){
			for(int j=1; j<dimY-1; j++){
				for(int k=1; k<dimX-1; k++){
					double total_w = 0;
					float v[3];
					v[0] = v[1] = v[2] = 0;
					for(int z=-1; z<2; z++){
						for(int y=-1; y<2; y++){
							for(int x=-1; x<2; x++){
								int kx = k + x;
								int jy = j + y;
								int iz = i + z;
								double w = spaceX*(value[iz][jy][kx+1] + value[iz][jy][kx-1])
									+ spaceY*(value[iz][jy+1][kx] + value[iz][jy-1][kx])
									+ spaceZ*(value[iz+1][jy][kx] + value[iz-1][jy][kx])
									- 2.0*(spaceX+spaceY+spaceZ)*value[iz][jy][kx];
								w = exp(-T*fabs(w));
								total_w += w;
								v[0] += w*nabla[iz][jy][kx][0];
								v[1] += w*nabla[iz][jy][kx][1];
								v[2] += w*nabla[iz][jy][kx][2];
							}
						}
					}
					if(total_w != 0){
						tmp[i][j][k][0] = v[0]/total_w;
						tmp[i][j][k][1] = v[1]/total_w;
						tmp[i][j][k][2] = v[2]/total_w;
					}
					else{
						tmp[i][j][k][0] = nabla[i][j][k][0];
						tmp[i][j][k][1] = nabla[i][j][k][1];
						tmp[i][j][k][2] = nabla[i][j][k][2];
					}
				}
			}
		}
		for(i=1; i<dimZ-1; i++){
			for(int j=1; j<dimY-1; j++){
				for(int k=1; k<dimX-1; k++){
					nabla[i][j][k][0] = tmp[i][j][k][0];
					nabla[i][j][k][1] = tmp[i][j][k][1];
					nabla[i][j][k][2] = tmp[i][j][k][2];
				}
			}
		}
	}

	for(i=0; i<dimZ; i++){
		for(int j=0; j<dimY; j++)
			delete[] tmp[i][j];
		delete[] tmp[i];
	}
	delete[] tmp;
}

void VolumeFunction::averageGradient2(int iter, float T)
{
	float (***tmp)[3] = (float(***)[3])new float**[dimZ];
	for(int i=0; i<dimZ; i++){
		tmp[i] = (float(**)[3])new float*[dimY];
		for(int j=0; j<dimY; j++){
			tmp[i][j] = new float[dimX][3];
		}
	}

	for(int m=0; m<iter; m++){
		for(i=2; i<dimZ-2; i++){
			for(int j=2; j<dimY-2; j++){
				for(int k=2; k<dimX-2; k++){
					double total_w = 0;
					double v[3];
					v[0] = v[1] = v[2] = 0;
					for(int z=-1; z<2; z++){
						for(int y=-1; y<2; y++){
							for(int x=-1; x<2; x++){
								int kx = k + x;
								int jy = j + y;
								int iz = i + z;
								double w = spaceX*(nabla[iz][jy][kx+1][0] - nabla[iz][jy][kx-1][0]) +
									spaceY*(nabla[iz][jy+1][kx][1] - nabla[iz][jy-1][kx][1]) +
									spaceZ*(nabla[iz+1][jy][kx][2] - nabla[iz-1][jy][kx][2]);
								w = exp(-T*fabs(w));
								total_w += w;
								v[0] += w*nabla[iz][jy][kx][0];
								v[1] += w*nabla[iz][jy][kx][1];
								v[2] += w*nabla[iz][jy][kx][2];
							}
						}
					}
					if((float)total_w != 0){
						tmp[i][j][k][0] = v[0]/total_w;
						tmp[i][j][k][1] = v[1]/total_w;
						tmp[i][j][k][2] = v[2]/total_w;
					}
					else{
						tmp[i][j][k][0] = nabla[i][j][k][0];
						tmp[i][j][k][1] = nabla[i][j][k][1];
						tmp[i][j][k][2] = nabla[i][j][k][2];
					}
				}
			}
		}
		for(i=2; i<dimZ-2; i++){
			for(int j=2; j<dimY-2; j++){
				for(int k=2; k<dimX-2; k++){
					nabla[i][j][k][0] = 0.5f*nabla[i][j][k][0] + 0.5f*tmp[i][j][k][0];
					nabla[i][j][k][1] = 0.5f*nabla[i][j][k][1] + 0.5f*tmp[i][j][k][1];
					nabla[i][j][k][2] = 0.5f*nabla[i][j][k][2] + 0.5f*tmp[i][j][k][2];
				}
			}
		}
	}

	for(i=0; i<dimZ; i++){
		for(int j=0; j<dimY; j++)
			delete[] tmp[i][j];
		delete[] tmp[i];
	}
	delete[] tmp;
}

void VolumeFunction::averageGradient3(int iter, float T)
{
	float (***tmp)[3] = (float(***)[3])new float**[dimZ];
	for(int i=0; i<dimZ; i++){
		tmp[i] = (float(**)[3])new float*[dimY];
		for(int j=0; j<dimY; j++){
			tmp[i][j] = new float[dimX][3];
		}
	}

	for(int m=0; m<iter; m++){
		for(i=2; i<dimZ-2; i++){
			for(int j=2; j<dimY-2; j++){
				for(int k=2; k<dimX-2; k++){
					double total_w = 0;
					float v[3], v0[3];
					v[0] = v[1] = v[2] = 0;
					double len = MeshData::LENGTH(nabla[i][j][k]);
					if((float)len == 0)
						continue;
					v0[0] = nabla[i][j][k][0]/len;
					v0[1] = nabla[i][j][k][1]/len;
					v0[2] = nabla[i][j][k][2]/len;
					for(int z=-1; z<2; z++){
						for(int y=-1; y<2; y++){
							for(int x=-1; x<2; x++){
								float* v1 = nabla[i+z][j+y][k+x];
								double len1 = MeshData::LENGTH(v1);
								if((float)len1 == 0)
									continue;
								int d2 = abs(x) + abs(y) + abs(z) + 1;
								double k = acos(MeshData::DOT(v0, v1)/len1);
								double w = exp(-T*k*k/d2);
								total_w += w;
								v[0] += w*v1[0];
								v[1] += w*v1[1];
								v[2] += w*v1[2];
							}
						}
					}
					if((float)total_w != 0){
						tmp[i][j][k][0] = v[0]/total_w;
						tmp[i][j][k][1] = v[1]/total_w;
						tmp[i][j][k][2] = v[2]/total_w;
					}
					else{
						tmp[i][j][k][0] = nabla[i][j][k][0];
						tmp[i][j][k][1] = nabla[i][j][k][1];
						tmp[i][j][k][2] = nabla[i][j][k][2];
					}
				}
			}
		}
		for(i=2; i<dimZ-2; i++){
			for(int j=2; j<dimY-2; j++){
				for(int k=2; k<dimX-2; k++){
					nabla[i][j][k][0] = tmp[i][j][k][0];
					nabla[i][j][k][1] = tmp[i][j][k][1];
					nabla[i][j][k][2] = tmp[i][j][k][2];
				}
			}
		}
	}

	for(i=0; i<dimZ; i++){
		for(int j=0; j<dimY; j++)
			delete[] tmp[i][j];
		delete[] tmp[i];
	}
	delete[] tmp;
}

void VolumeFunction::generateDist(int max_dist)
{
	dist = new int**[dimZ];
	for(int i=0; i<dimZ; i++){
		dist[i] = new int*[dimY];
		for(int j=0; j<dimY; j++)
			dist[i][j] = new int[dimX];
	}

	for(i=0; i<dimZ; i++){
		for(int j=0; j<dimY; j++){
			for(int k=0; k<dimX; k++){
			}
		}
	}
}	

MeshData* VolumeFunction::generateMesh()
{
	int ***index = new int**[dimZ+1];
	for(int i=0; i<dimZ+1; i++){
		index[i] = new int*[dimY+1];
		for(int j=0; j<dimY+1; j++){
			index[i][j] = new int[dimX+1];
			for(int k=0; k<dimX+1; k++)
				index[i][j][k] = -1;
		}
	}

	int current = 0;
	int face_N = 0;
	for(i=0; i<dimZ; i++)
		for(int j=0; j<dimY; j++)
			for(int k=0; k<dimX-1; k++){
				if(value[i][j][k] >= 0 && value[i][j][k+1] < 0){
					face_N++;
					if(index[i][j][k+1] < 0){
						index[i][j][k+1] = current;
						current++;
					}
					if(index[i][j+1][k+1] < 0){
						index[i][j+1][k+1] = current;
						current++;
					}
					if(index[i+1][j][k+1] < 0){
						index[i+1][j][k+1] = current;
						current++;
					}
					if(index[i+1][j+1][k+1] < 0){
						index[i+1][j+1][k+1] = current;
						current++;
					}
				}
				else if(value[i][j][k] < 0 && value[i][j][k+1] >= 0){
					face_N++;
					if(index[i][j][k+1] < 0){
						index[i][j][k+1] = current;
						current++;
					}
					if(index[i][j+1][k+1] < 0){
						index[i][j+1][k+1] = current;
						current++;
					}
					if(index[i+1][j][k+1] < 0){
						index[i+1][j][k+1] = current;
						current++;
					}
					if(index[i+1][j+1][k+1] < 0){
						index[i+1][j+1][k+1] = current;
						current++;
					}
				}
			}

	for(i=0; i<dimX; i++)
		for(int j=0; j<dimZ; j++)
			for(int k=0; k<dimY-1; k++){
				if(value[j][k][i] >= 0 && value[j][k+1][i] < 0){
					face_N++;
					if(index[j][k+1][i] < 0){
						index[j][k+1][i] = current;
						current++;
					}
					if(index[j+1][k+1][i] < 0){
						index[j+1][k+1][i] = current;
						current++;
					}
					if(index[j][k+1][i+1] < 0){
						index[j][k+1][i+1] = current;
						current++;
					}
					if(index[j+1][k+1][i+1] < 0){
						index[j+1][k+1][i+1] = current;
						current++;
					}
				}
				else if(value[j][k][i] < 0 && value[j][k+1][i] >= 0){
					face_N++;
					if(index[j][k+1][i] < 0){
						index[j][k+1][i] = current;
						current++;
					}
					if(index[j+1][k+1][i] < 0){
						index[j+1][k+1][i] = current;
						current++;
					}
					if(index[j][k+1][i+1] < 0){
						index[j][k+1][i+1] = current;
						current++;
					}
					if(index[j+1][k+1][i+1] < 0){
						index[j+1][k+1][i+1] = current;
						current++;
					}
				}
			}

	for(i=0; i<dimY; i++)
		for(int j=0; j<dimX; j++)
			for(int k=0; k<dimZ-1; k++){
				if(value[k][i][j] >= 0 && value[k+1][i][j] < 0){
					face_N++;
					if(index[k+1][i][j] < 0){
						index[k+1][i][j] = current;
						current++;
					}
					if(index[k+1][i+1][j] < 0){
						index[k+1][i+1][j] = current;
						current++;
					}
					if(index[k+1][i][j+1] < 0){
						index[k+1][i][j+1] = current;
						current++;
					}
					if(index[k+1][i+1][j+1] < 0){
						index[k+1][i+1][j+1] = current;
						current++;
					}
				}
				else if(value[k][i][j] < 0 && value[k+1][i][j] >= 0){
					face_N++;
					if(index[k+1][i][j] < 0){
						index[k+1][i][j] = current;
						current++;
					}
					if(index[k+1][i+1][j] < 0){
						index[k+1][i+1][j] = current;
						current++;
					}
					if(index[k+1][i][j+1] < 0){
						index[k+1][i][j+1] = current;
						current++;
					}
					if(index[k+1][i+1][j+1] < 0){
						index[k+1][i+1][j+1] = current;
						current++;
					}
				}
			}
		
	MeshData* mesh = new MeshData;
	mesh->setVertexCount(current);
	float (*vertex)[3] = mesh->vertex;
	current = 0;
	for(i=0; i<dimZ+1; i++)
		for(int j=0; j<dimY+1; j++)
			for(int k=0; k<dimX+1; k++)
				if(index[i][j][k] >= 0){
					index[i][j][k] = current;
					vertex[current][0] = spaceX*((float)k-0.5f) + originX;
					vertex[current][1] = spaceY*((float)j-0.5f) + originY;
					vertex[current][2] = spaceZ*((float)i-0.5f) + originZ;
					current++;
				}

	mesh->setFaceCount(2*face_N);

	face_N = 0;
	int (*face)[3] = mesh->face;
	for(i=0; i<dimZ; i++)
		for(int j=0; j<dimY; j++)
			for(int k=0; k<dimX-1; k++)
				if(value[i][j][k] >= 0 && value[i][j][k+1] < 0){
					face[face_N][0] = index[i][j][k+1];
					face[face_N][1] = index[i][j+1][k+1];
					face[face_N][2] = index[i+1][j][k+1];
					face_N++;

					face[face_N][0] = index[i][j+1][k+1];
					face[face_N][1] = index[i+1][j+1][k+1];
					face[face_N][2] = index[i+1][j][k+1];
					face_N++;
				}
				else if(value[i][j][k] < 0 && value[i][j][k+1] >= 0){
					face[face_N][0] = index[i][j][k+1];
					face[face_N][1] = index[i+1][j][k+1];
					face[face_N][2] = index[i+1][j+1][k+1];
					face_N++;

					face[face_N][0] = index[i][j+1][k+1];
					face[face_N][1] = index[i][j][k+1];
					face[face_N][2] = index[i+1][j+1][k+1];
					face_N++;
				}

	for(i=0; i<dimX; i++)
		for(int j=0; j<dimZ; j++)
			for(int k=0; k<dimY-1; k++)
				if(value[j][k][i] < 0 && value[j][k+1][i] >= 0){
					face[face_N][0] = index[j][k+1][i];
					face[face_N][2] = index[j+1][k+1][i];
					face[face_N][1] = index[j][k+1][i+1];
					face_N++;

					face[face_N][0] = index[j+1][k+1][i];
					face[face_N][2] = index[j+1][k+1][i+1];
					face[face_N][1] = index[j][k+1][i+1];
					face_N++;
				}
				else if(value[j][k][i] >= 0 && value[j][k+1][i] < 0){
					face[face_N][0] = index[j][k+1][i];
					face[face_N][2] = index[j][k+1][i+1];
					face[face_N][1] = index[j+1][k+1][i+1];
					face_N++;

					face[face_N][0] = index[j+1][k+1][i];
					face[face_N][2] = index[j][k+1][i];
					face[face_N][1] = index[j+1][k+1][i+1];
					face_N++;
				}

	for(i=0; i<dimY; i++)
		for(int j=0; j<dimX; j++)
			for(int k=0; k<dimZ-1; k++)
				if(value[k][i][j] >= 0 && value[k+1][i][j] < 0){
					face[face_N][0] = index[k+1][i][j];
					face[face_N][1] = index[k+1][i][j+1];
					face[face_N][2] = index[k+1][i+1][j];
					face_N++;

					face[face_N][0] = index[k+1][i][j+1];
					face[face_N][1] = index[k+1][i+1][j+1];
					face[face_N][2] = index[k+1][i+1][j];
					face_N++;
				}
				else if(value[k][i][j] < 0 && value[k+1][i][j] >= 0){
					face[face_N][0] = index[k+1][i][j];
					face[face_N][1] = index[k+1][i+1][j];
					face[face_N][2] = index[k+1][i+1][j+1];
					face_N++;

					face[face_N][0] = index[k+1][i][j+1];
					face[face_N][1] = index[k+1][i][j];
					face[face_N][2] = index[k+1][i+1][j+1];
					face_N++;
				}

	for(i=0; i<dimZ+1; i++){
		for(int j=0; j<dimY+1; j++)
			delete[] index[i][j];
		delete[] index[i];
	}
	delete[] index;

	return mesh;
}

void VolumeFunction::sampleImplicitFunction(vtkImplicitFunction *func)
{
	nabla = (float(***)[3])new float**[dimZ];
	for(int i=0; i<dimZ; i++){
		nabla[i] = (float(**)[3])new float*[dimY];
		for(int j=0; j<dimY; j++){
			nabla[i][j] = new float[dimX][3];
		}
	}

	float p[3];
	for(i=0; i<dimZ; i++){
		p[2] = originZ + i*spaceZ;
		for(int j=0; j<dimY; j++){
			p[1] = originY + j*spaceY;
			for(int k=0; k<dimX; k++){
				p[0] = originX + k*spaceX;
				value[i][j][k] = func->EvaluateFunction(p);
				func->EvaluateGradient(p, nabla[i][j][k]);
			}
		}
	}
}

void VolumeFunction::sampleImplicitValue(vtkImplicitFunction *func)
{
	float p[3];
	for(int i=0; i<dimZ; i++){
		p[2] = originZ + i*spaceZ;
		for(int j=0; j<dimY; j++){
			p[1] = originY + j*spaceY;
			for(int k=0; k<dimX; k++){
				p[0] = originX + k*spaceX;
				value[i][j][k] = func->EvaluateFunction(p);
			}
		}
	}
}

MeshData* VolumeFunction::generateMesh(int sh)
{
	int dimZ_o = dimZ;
	int dimY_o = dimY;
	int dimX_o = dimX;
	float spaceZ_o = spaceZ;
	float spaceY_o = spaceY;
	float spaceX_o = spaceX;
	float*** value_o = value;

	dimZ = dimZ/sh*0.4;
	dimY = dimY/sh;
	dimX = dimX/sh;

	spaceX *= sh;
	spaceY *= sh;
	spaceZ *= sh;

	int i;

	
	value = new float**[dimZ];
	for(i=0; i<dimZ; i++){
		value[i] = new float*[dimY];
		for(int j=0; j<dimY; j++){
			value[i][j] = new float[dimX];
		}
	}

	for(i=0; i<dimZ; i++){
		for(int j=0; j<dimY; j++){
			for(int k=0; k<dimX; k++){
				value[i][j][k] = 0;
				for(int z=0; z<sh; z++)
					for(int y=0; y<sh; y++)
						for(int x=0; x<sh; x++)
							value[i][j][k] += value_o[sh*i+z][sh*j+y][sh*k+x];
				value[i][j][k] /= pow(sh,3);
			}
		}
	}

	int ***index = new int**[dimZ+1];
	for(i=0; i<dimZ+1; i++){
		index[i] = new int*[dimY+1];
		for(int j=0; j<dimY+1; j++){
			index[i][j] = new int[dimX+1];
			for(int k=0; k<dimX+1; k++)
				index[i][j][k] = -1;
		}
	}

	int current = 0;
	int face_N = 0;
	for(i=0; i<dimZ; i++)
		for(int j=0; j<dimY; j++)
			for(int k=0; k<dimX-1; k++){
				if(value[i][j][k] >= 0 && value[i][j][k+1] < 0){
					face_N++;
					if(index[i][j][k+1] < 0){
						index[i][j][k+1] = current;
						current++;
					}
					if(index[i][j+1][k+1] < 0){
						index[i][j+1][k+1] = current;
						current++;
					}
					if(index[i+1][j][k+1] < 0){
						index[i+1][j][k+1] = current;
						current++;
					}
					if(index[i+1][j+1][k+1] < 0){
						index[i+1][j+1][k+1] = current;
						current++;
					}
				}
				else if(value[i][j][k] < 0 && value[i][j][k+1] >= 0){
					face_N++;
					if(index[i][j][k+1] < 0){
						index[i][j][k+1] = current;
						current++;
					}
					if(index[i][j+1][k+1] < 0){
						index[i][j+1][k+1] = current;
						current++;
					}
					if(index[i+1][j][k+1] < 0){
						index[i+1][j][k+1] = current;
						current++;
					}
					if(index[i+1][j+1][k+1] < 0){
						index[i+1][j+1][k+1] = current;
						current++;
					}
				}
			}

	for(i=0; i<dimX; i++)
		for(int j=0; j<dimZ; j++)
			for(int k=0; k<dimY-1; k++){
				if(value[j][k][i] >= 0 && value[j][k+1][i] < 0){
					face_N++;
					if(index[j][k+1][i] < 0){
						index[j][k+1][i] = current;
						current++;
					}
					if(index[j+1][k+1][i] < 0){
						index[j+1][k+1][i] = current;
						current++;
					}
					if(index[j][k+1][i+1] < 0){
						index[j][k+1][i+1] = current;
						current++;
					}
					if(index[j+1][k+1][i+1] < 0){
						index[j+1][k+1][i+1] = current;
						current++;
					}
				}
				else if(value[j][k][i] < 0 && value[j][k+1][i] >= 0){
					face_N++;
					if(index[j][k+1][i] < 0){
						index[j][k+1][i] = current;
						current++;
					}
					if(index[j+1][k+1][i] < 0){
						index[j+1][k+1][i] = current;
						current++;
					}
					if(index[j][k+1][i+1] < 0){
						index[j][k+1][i+1] = current;
						current++;
					}
					if(index[j+1][k+1][i+1] < 0){
						index[j+1][k+1][i+1] = current;
						current++;
					}
				}
			}

	for(i=0; i<dimY; i++)
		for(int j=0; j<dimX; j++)
			for(int k=0; k<dimZ-1; k++){
				if(value[k][i][j] >= 0 && value[k+1][i][j] < 0){
					face_N++;
					if(index[k+1][i][j] < 0){
						index[k+1][i][j] = current;
						current++;
					}
					if(index[k+1][i+1][j] < 0){
						index[k+1][i+1][j] = current;
						current++;
					}
					if(index[k+1][i][j+1] < 0){
						index[k+1][i][j+1] = current;
						current++;
					}
					if(index[k+1][i+1][j+1] < 0){
						index[k+1][i+1][j+1] = current;
						current++;
					}
				}
				else if(value[k][i][j] < 0 && value[k+1][i][j] >= 0){
					face_N++;
					if(index[k+1][i][j] < 0){
						index[k+1][i][j] = current;
						current++;
					}
					if(index[k+1][i+1][j] < 0){
						index[k+1][i+1][j] = current;
						current++;
					}
					if(index[k+1][i][j+1] < 0){
						index[k+1][i][j+1] = current;
						current++;
					}
					if(index[k+1][i+1][j+1] < 0){
						index[k+1][i+1][j+1] = current;
						current++;
					}
				}
			}
		
	MeshData* mesh = new MeshData;
	mesh->setVertexCount(current);
	float (*vertex)[3] = mesh->vertex;
	current = 0;
	for(i=0; i<dimZ+1; i++)
		for(int j=0; j<dimY+1; j++)
			for(int k=0; k<dimX+1; k++)
				if(index[i][j][k] >= 0){
					index[i][j][k] = current;
					vertex[current][0] = spaceX*((float)k-0.5f) + originX;
					vertex[current][1] = spaceY*((float)j-0.5f) + originY;
					vertex[current][2] = spaceZ*((float)i-0.5f) + originZ;
					current++;
				}

	mesh->setFaceCount(2*face_N);

	face_N = 0;
	int (*face)[3] = mesh->face;
	for(i=0; i<dimZ; i++)
		for(int j=0; j<dimY; j++)
			for(int k=0; k<dimX-1; k++)
				if(value[i][j][k] >= 0 && value[i][j][k+1] < 0){
					face[face_N][0] = index[i][j][k+1];
					face[face_N][1] = index[i][j+1][k+1];
					face[face_N][2] = index[i+1][j][k+1];
					face_N++;

					face[face_N][0] = index[i][j+1][k+1];
					face[face_N][1] = index[i+1][j+1][k+1];
					face[face_N][2] = index[i+1][j][k+1];
					face_N++;
				}
				else if(value[i][j][k] < 0 && value[i][j][k+1] >= 0){
					face[face_N][0] = index[i][j][k+1];
					face[face_N][1] = index[i+1][j][k+1];
					face[face_N][2] = index[i+1][j+1][k+1];
					face_N++;

					face[face_N][0] = index[i][j+1][k+1];
					face[face_N][1] = index[i][j][k+1];
					face[face_N][2] = index[i+1][j+1][k+1];
					face_N++;
				}

	for(i=0; i<dimX; i++)
		for(int j=0; j<dimZ; j++)
			for(int k=0; k<dimY-1; k++)
				if(value[j][k][i] < 0 && value[j][k+1][i] >= 0){
					face[face_N][0] = index[j][k+1][i];
					face[face_N][2] = index[j+1][k+1][i];
					face[face_N][1] = index[j][k+1][i+1];
					face_N++;

					face[face_N][0] = index[j+1][k+1][i];
					face[face_N][2] = index[j+1][k+1][i+1];
					face[face_N][1] = index[j][k+1][i+1];
					face_N++;
				}
				else if(value[j][k][i] >= 0 && value[j][k+1][i] < 0){
					face[face_N][0] = index[j][k+1][i];
					face[face_N][2] = index[j][k+1][i+1];
					face[face_N][1] = index[j+1][k+1][i+1];
					face_N++;

					face[face_N][0] = index[j+1][k+1][i];
					face[face_N][2] = index[j][k+1][i];
					face[face_N][1] = index[j+1][k+1][i+1];
					face_N++;
				}

	for(i=0; i<dimY; i++)
		for(int j=0; j<dimX; j++)
			for(int k=0; k<dimZ-1; k++)
				if(value[k][i][j] >= 0 && value[k+1][i][j] < 0){
					face[face_N][0] = index[k+1][i][j];
					face[face_N][1] = index[k+1][i][j+1];
					face[face_N][2] = index[k+1][i+1][j];
					face_N++;

					face[face_N][0] = index[k+1][i][j+1];
					face[face_N][1] = index[k+1][i+1][j+1];
					face[face_N][2] = index[k+1][i+1][j];
					face_N++;
				}
				else if(value[k][i][j] < 0 && value[k+1][i][j] >= 0){
					face[face_N][0] = index[k+1][i][j];
					face[face_N][1] = index[k+1][i+1][j];
					face[face_N][2] = index[k+1][i+1][j+1];
					face_N++;

					face[face_N][0] = index[k+1][i][j+1];
					face[face_N][1] = index[k+1][i][j];
					face[face_N][2] = index[k+1][i+1][j+1];
					face_N++;
				}

	for(i=0; i<dimZ+1; i++){
		for(int j=0; j<dimY+1; j++)
			delete[] index[i][j];
		delete[] index[i];
	}
	delete[] index;


	dimZ = dimZ_o;
	dimY = dimY_o;
	dimX = dimX_o;

	spaceZ = spaceZ_o;
	spaceY = spaceY_o;
	spaceX = spaceX_o;

	value = value_o;

	return mesh;
}
